/*  Python extension module for Kerberos 5 interfaces.
 *
 *  Two supplemental sources are used to implement the TCP and UDP versions
 *  of the auth_context interface; they really are quite different.
 *
 *	krb5tcp.c	-- the TCP version
 *	krb5udp.c	-- the UDP version
 *
 *  This source file includes the implementation of the Principle object and
 *  the module initialization function.
 *
 *  Written by Fred L. Drake, Jr.  <fdrake@cnri.reston.va.us>
 *	      and Roger E. Masse   <rmasse@cnri.reston.va.us>
 */


#include "krb5py.h"			/* common header file */
#include "structmember.h"		/* Python attribute lookup */

#include <netdb.h>			/* gethostbyname(), struct hostent */
#include <unistd.h>			/* gethostname() */
#include <sys/socket.h>			/* getpeername() */


/*  Global krb5 module variables.
 *
 *  These are the context and credentials cache used by the module, plus
 *  an exception which can be raised to alert the Python layer of errors.
 */
krb5_context pk_context;
krb5_ccache pk_ccache;

PyObject *pk_exception;


static void
finish_error(rc, msg)
    krb5_error_code rc;
    char *msg;
{
    PyObject *py_rc = PyInt_FromLong(rc);
    PyObject *py_msg = PyString_FromString(msg);

    if ((py_rc != NULL) && (py_msg != NULL)) {
	if (PyClass_Check(pk_exception)) {
	    PyObject *exc = PyObject_CallFunction(
		pk_exception, "OO", py_rc, py_msg);

	    if (exc != NULL) {
		PyObject_SetAttrString(exc, "err_code", py_rc);
		PyObject_SetAttrString(exc, "message", py_msg);
		PyErr_SetObject(pk_exception, exc);
	    }
	}
	else
	    PyErr_SetObject(pk_exception, Py_BuildValue("OO", py_rc, py_msg));
    }
    Py_XDECREF(py_rc);
    Py_XDECREF(py_msg);

}   /* finish_error() */


void
pk_error(rc)
    krb5_error_code rc;
{
    if (rc == ENOMEM) {
	/* Lots of krb5_*() routines engage in memory allocation
	 * and can return this, based on the documentation.
	 */
	PyErr_NoMemory();
	return;
    }
#if 0
    if (rc == ASN1_MISSING_FIELD) {
	extern char *evil_debug_value;
	fprintf(stderr,
		"ASN1_MISSING_FIELD raised by %s\n",
		evil_debug_value ? evil_debug_value : "(NULL)");
    }
#endif
    finish_error(rc, error_message(rc));

}   /* pk_error() */


void pk_var_error(krb5_error_code rc, const char *fmt, ...)
{
    static char errbuf[1024];
    va_list args;

    *errbuf = '\0';
    if (rc) {
	strcat(errbuf, error_message(rc));
	strcat(errbuf, " ");
    }
    if (args)
        vsprintf(errbuf + strlen (errbuf), fmt, args);
    else
        strcat(errbuf, fmt);

    finish_error(rc, errbuf);

}   /* pk_var_error() */


static PyObject*
get_socket_error()
{
    static PyObject *socket_error = NULL;

    if (socket_error == NULL) {
	PyObject *module = PyImport_ImportModule("socket");

	if (module != NULL)
	    socket_error = PyObject_GetAttrString(module, "error");
    }
    return (socket_error);

}   /* get_socket_error() */


void
pk_socket_error()
{
    PyObject *error = get_socket_error();

#ifdef WIN32
    /* This is boilerplate & hasn't been tested as part of this module.
     * Good luck.
     */
    char *buffer;
    int errcode = WSAGetLastError();
    DWORD flags = (FORMAT_MESSAGE_ALLOCATE_BUFFER
		   | FORMAT_MESSAGE_FROM_SYSTEM
		   | FORMAT_MESSAGE_IGNORE_INSERTS);
    int len = FormatMessage(flags, NULL, errcode,
			    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			    (LPTSTR)&buffer, 0, NULL);

    if (len == 0)
	PyErr_SetObject(error,
			Py_BuildValue("is", errcode, "winsock error"));
    else
	PyErr_SetObject(error,
			Py_BuildValue("is#", errcode, buffer, len));
    LocalFree(buffer);
#else
    PyErr_SetFromErrno(error);
#endif

}   /* pk_socket_error() */


unsigned long
pk_parse_address(address)
    const char *address;
{
    static unsigned long local_ip = 0;

    int d1, d2, d3, d4;
    char ch;
    char hostname[1024];
    int is_local = 0;
    unsigned long result = 0;

    if (address[0] == (char)0) {
	/* use the local address */
	if (local_ip == 0) {
	    if (gethostname(hostname, sizeof(hostname) - 1)) {
		pk_socket_error();
		return (0);
	    }
	    address = hostname;
	    is_local = 1;
	}
	else
	    return (local_ip);
    }
    if ((sscanf(address, "%d.%d.%d.%d%c", &d1, &d2, &d3, &d4, &ch) != 4)
	|| (0 > d1) || (d1 > 255) || (0 > d2) || (d2 > 255)
	|| (0 > d3) || (d3 > 255) || (0 > d4) || (d4 > 255)) {
	/* we have a host name */
	struct hostent *host;

	host = gethostbyname(address);
	if (host == NULL)
	    pk_socket_error();
	else {
	    memcpy(&result, host->h_addr, sizeof(result));
	    if (is_local)
		local_ip = result;
	}
    }
    else
	/* we have an IP number */
	result = (unsigned long)
	    htonl((d1 << 24) | (d2 << 16) | (d3 << 8) | d4);

    return (result);

}   /* pk_parse_address() */


static PyObject*
princ_name(self, args)
    Principal *self;
    PyObject *args;
{
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, ":name")) {
	krb5_error_code rc;
	char *name;

	if (self->name == NULL) {
	    rc = krb5_unparse_name(pk_context, self->principal, &name);
	    if (rc)
		pk_error(rc);
	    else {
		self->name = PyString_FromString(name);
		free(name);
	    }
	}
	result = self->name;
	Py_XINCREF(result);
    }
    return (result);

}   /* princ_name() */


static PyObject*
princ_realm(self, args)
    Principal *self;
    PyObject *args;
{
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, ":realm")) {
	krb5_data *realm;

	realm = krb5_princ_realm(pk_context, self->principal);
	result = PyString_FromStringAndSize(realm->data, realm->length);
    }
    return (result);

}   /* princ_realm() */


static PyMethodDef
princ_methods[] = {
    {"name",		princ_name,		METH_VARARGS,
     "Retrieve the full name of the principal, including the realm."},
    {"realm",		princ_realm,		METH_VARARGS,
     "Retrieve the realm of the principal."},
    {NULL}

};   /* princ_methods[] */


#define OFF(x) offsetof(Principal, x)

static struct memberlist
princ_memberlist[] = {
    {"host",		T_OBJECT,	OFF(host),		RO},
    {"service",		T_OBJECT,	OFF(service),		RO},

    {NULL}
};


static PyObject*
princ_getattr(self, name)
    PyObject *self;
    char *name;
{
    PyObject *result = PyMember_Get((char *)self, princ_memberlist, name);

    if (result == NULL) {
	PyErr_Clear();
	result = Py_FindMethod(princ_methods, self, name);
    }
    return (result);

}   /* princ_getattr() */


static PyObject*
empty_args_list = NULL;

static PyObject*
princ_repr(self)
    Principal *self;
{
    PyObject *name = princ_name(self, empty_args_list);
    PyObject *result = NULL;

    if (name != NULL) {
	char buffer[256];

	sprintf(buffer, "<krb5.Principal %s>", PyString_AS_STRING(name));
	Py_DECREF(name);
	result = PyString_FromString(buffer);
    }
    return (result);

}   /* princ_repr() */


static int
princ_compare(self, other)
    Principal *self;
    PyObject *other;
{
    if (PyObject_Type(other) != (PyObject *)&pk_PrincipalType)
	return (-1);

    return (!krb5_principal_compare(
	pk_context, self->principal, ((Principal *)other)->principal));

}   /* princ_compare() */


static void
princ_free(self)
    Principal *self;
{
    Py_XDECREF(self->host);
    Py_XDECREF(self->service);
    Py_XDECREF(self->name);
    krb5_free_principal(pk_context, self->principal);
    PyMem_DEL(self);

}   /* princ_free() */


PyTypeObject
pk_PrincipalType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,
    "Principal",			/* tp_name		*/
    sizeof(Principal),			/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    (destructor)princ_free,		/* tp_dealloc		*/
    0,					/* tp_print		*/
    princ_getattr,			/* tp_getattr		*/
    0,					/* tp_setattr		*/
    princ_compare,			/* tp_compare		*/
    princ_repr,				/* tp_repr		*/
    0,					/* tp_as_number		*/
    0,					/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    0,					/* tp_getattro		*/
    0,					/* tp_setattro		*/

    /* Functions to access object as input/output buffer */
    0,					/* tp_as_buffer		*/

    /* Space for future expansion */
    0,					/* tp_xxx4		*/

    /* __doc__ */
    "Kerberos principal."
};


Principal*
pk_new_Principal(hostname, service)
    PyObject *hostname;
    PyObject *service;
{
    Principal *princ = PyObject_NEW(Principal, &pk_PrincipalType);

    if (princ != NULL) {
	Py_INCREF(princ->host = hostname);
	Py_INCREF(princ->service = service);
	princ->name = NULL;
    }
    return (princ);

}   /* pk_new_Principal() */


static PyObject*
get_login_principal(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, ":get_login_principal")) {
	Principal *princ = pk_new_Principal(Py_None, Py_None);

	if (princ != NULL) {
	    krb5_error_code rc;

	    rc = krb5_cc_get_principal(
		pk_context, pk_ccache, &princ->principal);
	    if (rc) {
		Py_DECREF(princ->host);
		Py_DECREF(princ->service);
		PyMem_DEL(princ);
		pk_error(rc);
	    }
	    else
		result = (PyObject *)princ;
	}
    }
    return (result);

}   /* get_login_principal() */


static PyObject*
get_svc_principal(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    PyObject *hostname = NULL;
    PyObject *service;

    if (PyArg_ParseTuple(args, "S|S:get_svc_principal", &service, &hostname)) {
	char *svcname = PyString_AS_STRING(service);
	char *hname = NULL;
	char buffer[MAXHOSTNAMELEN];
	Principal *princ;
	krb5_error_code rc;
	struct hostent *hp;

	Py_XINCREF(hostname);
	if (hostname != NULL) {
	    int d1, d2, d3, d4;
	    char c;

	    hname = PyString_AS_STRING(hostname);
	    if (hname[0] == (char)0)
		hname = NULL;
	    else if (
		(sscanf(hname, "%d.%d.%d.%d%c", &d1, &d2, &d3, &d4, &c) == 4)
		&& (0 <  d1) && (d1 < 256) && (0 <= d2) && (d2 < 256)
		&& (0 <= d3) && (d3 < 256) && (0 <  d3) && (d3 < 256)) {
		/* numeric address, do a reverse lookup */
		struct hostent *hp;
		struct sockaddr_in addr;

		addr.sin_addr.s_addr = (unsigned long)
		    htonl((d1 << 24) | (d2 << 16) | (d3 << 8) | d4);
		hp = gethostbyaddr((char *)&addr.sin_addr,
				   sizeof(addr.sin_addr), AF_INET);
		if (hp == NULL) {
#ifdef HAVE_HSTRERROR
		    /* Let's get real error message to return */
		    extern int h_errno;
		    PyErr_SetString(get_socket_error(),
				    (char *)hstrerror(h_errno));
#else
		    PyErr_SetString(get_socket_error(), "host not found");
#endif
		    goto finally;
		}
		Py_DECREF(hostname);
		if ((hostname = PyString_FromString(hp->h_name)) == NULL)
		    goto finally;
		hname = PyString_AS_STRING(hostname);
	    }
	}
	if (hname == NULL) {
	    if (gethostname(buffer, sizeof(buffer))) {
		pk_socket_error();
		goto finally;
	    }
	    hname = buffer;
	}
	/* always get "official" host name */
	if ((hp = gethostbyname(hname)) == NULL) {
	    pk_socket_error();
	    goto finally;
	}
	Py_XDECREF(hostname);
	if ((hostname = PyString_FromString(hp->h_name)) == NULL)
	    goto finally;
	hname = PyString_AS_STRING(hostname);

	if (hostname == NULL)
	    Py_INCREF(hostname = Py_None);
	princ = pk_new_Principal(hostname, service);
	if (princ == NULL)
	    goto finally;
	rc = krb5_sname_to_principal(pk_context, hname, svcname,
				     KRB5_NT_SRV_HST, &(princ->principal));
	if (rc) {
	    pk_error(rc);
	    PyMem_DEL(princ);
	}
	else
	    result = (PyObject *)princ;
    }
  finally:
    Py_XDECREF(hostname);

    return (result);

}   /* get_svc_principal() */


static PyObject*
get_principal(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    Principal *princ = NULL;
    char *name;

    if (PyArg_ParseTuple(args, "s:get_principal", &name)) {
	princ = pk_new_Principal(Py_None, Py_None);
	if (princ != NULL) {
	    krb5_error_code rc;

	    rc = krb5_parse_name(pk_context, name, &princ->principal);
	    if (rc)
		pk_error(rc);
	    else
		Py_INCREF(result = (PyObject *)princ);
	}
    }
    Py_XDECREF(princ);

    return (result);

}   /* get_principal() */


/* convert socket arg to an int and insure that its connected */
int
pk_check_socket(sock)
    PyObject* sock;
{
    PyObject *meth;
    SOCKET s;
    struct sockaddr peername;
    int namelen = sizeof(peername);

    /* socket passed as an int */
    if (PyInt_Check(sock)) {
        s = PyInt_AsLong(sock);
    }

    /* socket passed as an object */
    else if ((meth = PyObject_GetAttrString(sock, "fileno")) != NULL) {
        PyObject *fno = PyEval_CallObject(meth, NULL);
        Py_DECREF(meth);
        if (fno == NULL)
	    return -1;

        if (!PyInt_Check(fno)) {
	    PyErr_SetString(PyExc_TypeError,
			    "fileno() method returned a non-integer.");
	    Py_DECREF(fno);
	    return -1;
         }
         s = PyInt_AsLong(fno);
         Py_DECREF(fno);
    }
    else {
         PyErr_SetString(PyExc_TypeError,
			 "socket must be an int or have a fileno() method.");
	 return -1;
    }
    /* insure that socket is connected */
    if (getpeername(s, &peername, &namelen)) {
	pk_socket_error();
	return -1;
    }
    return s;

}   /* pk_check_socket() */


int
pk_conv_fd(obj, result)
    PyObject *obj;
    void *result;
{
    int fd = pk_check_socket(obj);

    if (fd == -1)
	return (0);
    *((int *)result) = fd;

    return (1);

}   /* pk_conv_fd() */


static PyObject*
tcp_net_read(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    PyObject *readbuf = NULL;
    int size = 0;
    krb5_error_code rc;
    SOCKET s;

    if (PyArg_ParseTuple(args, "O&i:net_read", pk_conv_fd, &s, &size)) {
	if(! (readbuf = PyString_FromStringAndSize(NULL, size))) {
            PyErr_NoMemory();
            goto finally;
        }
        if ((rc = krb5_net_read(pk_context, s, PyString_AS_STRING(readbuf),
				size)) <= 0) {
            if (rc == 0)
		errno = ECONNABORTED;
	    pk_socket_error();
	    goto finally;
        }
	result = readbuf;
	readbuf = NULL;
    }
  finally:
    Py_XDECREF(readbuf);

    return (result);

}   /* tcp_net_read() */


static PyObject*
tcp_net_write(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    krb5_data userdata;
    krb5_error_code rc;
    SOCKET s;

    if (PyArg_ParseTuple(args, "O&s#:net_write", pk_conv_fd, &s,
			 &userdata.data, &userdata.length)) {
	if (!userdata.length) {
	    PyErr_SetString(PyExc_ValueError, "String must not be empty.");
	    goto finally;
	}
	if ((rc = krb5_net_write(pk_context, 0, (char *)userdata.data,
                                 userdata.length)) < 0) {
	    pk_socket_error();
            goto finally;
        }
        Py_INCREF(result = Py_None);
    }
  finally:
    return (result);

}   /* tcp_net_write() */


int
pk_tcp_setup_internals(self)
    AuthenticationContext *self;
{
    struct sockaddr_in address;
    struct sockaddr_in raddress;
    size_t sz = sizeof(address);

    if (self->have_rcache && self->have_addresses)
	return (0);
    if (getsockname(self->fd, (struct sockaddr *)&address, &sz)) {
	pk_socket_error();
	return (-1);
    }
    if (!self->have_addresses) {
	sz = sizeof(raddress);
	if (getpeername(self->fd, (struct sockaddr *)&raddress, &sz)) {
	    pk_socket_error();
	    return (-1);
	}
	if (pk_ac_set_addresses(self, &address, &raddress))
	    return (-1);
    }
    if (!self->have_rcache)
	return (pk_ac_set_rcache(self, &address, "_python_krb5_"));

    return (0);

}   /* pk_tcp_setup_internals() */


static PyObject*
ac_rd(self, args, argfmt, rd_func)
    AuthenticationContext *self;
    PyObject *args;
    char *argfmt;
    krb5_error_code (*rd_func)();
{
    PyObject *result = NULL;
    krb5_data inbuf;

    if (PyArg_ParseTuple(args, argfmt, &inbuf.data, &inbuf.length)) {
	krb5_error_code rc;
	krb5_data outbuf;

	if (self->fd > -1) {
	    /* if this is TCP, make sure everything is set up */
	    if (pk_tcp_setup_internals(self) == -1)
		goto finally;
	}
	rc = rd_func(pk_context, self->auth_context, &inbuf, &outbuf, NULL);
	if (rc)
	    pk_error(rc);
	else {
	    result = PyString_FromStringAndSize(outbuf.data, outbuf.length);
	    free(outbuf.data);
	}
    }
  finally:
    return (result);

}   /* ac_rd() */


PyObject*
pk_ac_rd_safe(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    return (ac_rd(self, args, "s#:rd_safe", krb5_rd_safe));

}   /* pk_ac_rd_safe() */


PyObject*
pk_ac_rd_priv(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    return (ac_rd(self, args, "s#:rd_priv", krb5_rd_priv));

}   /* pk_ac_rd_priv() */


static PyObject*
ac_mk(self, args, argfmt, mk_func)
    AuthenticationContext *self;
    PyObject *args;
    char *argfmt;
    krb5_error_code (*mk_func)();
{
    PyObject *result = NULL;
    krb5_data userdata;

    if (PyArg_ParseTuple(args, argfmt, &userdata.data, &userdata.length)) {
	krb5_data outbuf;
	krb5_error_code rc;

	if (self->fd > -1) {
	    /* if this is TCP, make sure everything is set up */
	    if (pk_tcp_setup_internals(self) == -1)
		goto finally;
	}
	outbuf.data = NULL;
	rc = mk_func(pk_context, self->auth_context, &userdata, &outbuf, NULL);
	if (rc)
	    pk_error(rc);
	else {
	    result = PyString_FromStringAndSize(outbuf.data, outbuf.length);
	    free(outbuf.data);
	}
    }
  finally:
    return (result);

}   /* ac_mk() */


PyObject*
pk_ac_mk_safe(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    return (ac_mk(self, args, "s#:mk_safe", krb5_mk_safe));

}   /* pk_ac_mk_safe() */


PyObject*
pk_ac_mk_priv(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    return (ac_mk(self, args, "s#:mk_priv", krb5_mk_priv));

}   /* pk_ac_mk_priv() */


PyObject*
pk_ac_getflags(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, ":getflags")) {
	krb5_int32 flags;
	krb5_error_code rc;

	rc = krb5_auth_con_getflags(pk_context, self->auth_context, &flags);
	if (rc)
	    pk_error(rc);
	else
	    result = PyInt_FromLong(flags);
    }
    return (result);

}   /* pk_ac_getflags() */


PyObject*
pk_ac_setflags(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    PyObject *result = NULL;
    int flags;

    if (PyArg_ParseTuple(args, "i:setflags", &flags)) {
	krb5_int32 legal = (KRB5_AUTH_CONTEXT_DO_TIME
			    | KRB5_AUTH_CONTEXT_RET_TIME
			    | KRB5_AUTH_CONTEXT_DO_SEQUENCE
			    | KRB5_AUTH_CONTEXT_RET_SEQUENCE);

	if (flags & ~legal)
	    PyErr_SetString(PyExc_ValueError, "illegal flag bits set");
	else {
	    krb5_error_code rc;

	    rc = krb5_auth_con_setflags(pk_context, self->auth_context, flags);
	    if (rc)
		pk_error(rc);
	    else {
		Py_INCREF(result = Py_None);
	    }
	}
    }
    return (result);

}   /* pk_ac_setflags() */


int
pk_ac_set_addresses(self, address, raddress)
    AuthenticationContext *self;
    struct sockaddr_in *address;
    struct sockaddr_in *raddress;
{
    krb5_error_code rc;
    krb5_address addr;
    krb5_address raddr;
    krb5_address *remote = NULL;
    int result = -1;

    if (raddress != NULL)
	remote = &raddr;

    /* make Kerberos calls */
    addr.addrtype = ADDRTYPE_IPPORT;
    addr.length = sizeof(address->sin_port);
    addr.contents = (krb5_octet *)&address->sin_port;
    if (remote != NULL) {
	raddr.addrtype = ADDRTYPE_IPPORT;
	raddr.length = sizeof(raddress->sin_port);
	raddr.contents = (krb5_octet *)&raddress->sin_port;
    }
    rc = krb5_auth_con_setports(pk_context, self->auth_context, &addr, remote);
    if (rc) {
	pk_error(rc);
	goto finally;
    }
    addr.addrtype = ADDRTYPE_INET;
    addr.length = sizeof(address->sin_addr);
    addr.contents = (krb5_octet *)&address->sin_addr;
    if (remote != NULL) {
	raddr.addrtype = ADDRTYPE_INET;
	raddr.length = sizeof(raddress->sin_addr);
	raddr.contents = (krb5_octet *)&raddress->sin_addr;
    }
    rc = krb5_auth_con_setaddrs(pk_context, self->auth_context, &addr, remote);
    if (rc)
	pk_error(rc);
    else {
	self->have_addresses = 1;
	result = 0;
    }
  finally:
    return (result);

}   /* pk_ac_set_addresses() */


int
pk_ac_set_rcache(self, address, cltname)
    AuthenticationContext *self;
    struct sockaddr_in *address;
    const char *cltname;
{
    int result = -1;
    krb5_address *outaddr = NULL;
    krb5_address addr;
    krb5_error_code rc;
    krb5_data piece;

    piece.data = NULL;
    piece.length = 0;
    addr.addrtype = ADDRTYPE_INET;
    addr.length = sizeof(address->sin_addr);
    addr.contents = (krb5_octet *)&address->sin_addr;
    rc = krb5_gen_portaddr(pk_context, &addr, &address->sin_port, &outaddr);
    if (rc) {
	pk_error(rc);
	goto finally;
    }
    rc = krb5_gen_replay_name(pk_context, outaddr, cltname, &piece.data);
    if (rc) {
	pk_error(rc);
	goto finally;
    }
    piece.length = strlen(piece.data);
    rc = krb5_get_server_rcache(pk_context, &piece, &(self->rcache));
    if (rc) {
	pk_error(rc);
	goto finally;
    }
    rc = krb5_auth_con_setrcache(pk_context, self->auth_context, self->rcache);
    if (rc)
	pk_error(rc);
    else {
	self->have_rcache = 1;
	result = 0;
    }
  finally:
    if (outaddr != NULL)
	krb5_free_address(pk_context, outaddr);
    free(piece.data);

    return (result);

}   /* pk_ac_set_rcache() */


static void
ac_free(self)
    AuthenticationContext *self;
{
    Py_XDECREF(self->local);
    Py_XDECREF(self->remote);
    Py_XDECREF(self->sock);
    if (self->ticket != NULL)
	krb5_free_ticket(pk_context, self->ticket);
    krb5_auth_con_free(pk_context, self->auth_context);
    PyMem_DEL(self);

}   /* ac_free() */


AuthenticationContext*
pk_new_AuthenticationContext(type)
    PyTypeObject *type;
{
    AuthenticationContext *acontext;

    if (type->tp_dealloc == NULL)
	type->tp_dealloc = ac_free;
    acontext = PyObject_NEW(AuthenticationContext, type);
    if (acontext != NULL) {
	krb5_error_code rc = krb5_auth_con_init(
	    pk_context, &(acontext->auth_context));

	if (rc) {
	    pk_error(rc);		/* raise exception */
	    PyMem_DEL(acontext);
	    acontext = NULL;
	}
	else {
	    acontext->fd = -1;
	    acontext->have_addresses = 0;
	    acontext->have_rcache = 0;
	    acontext->sock = NULL;
	    acontext->ticket = NULL;
	    Py_INCREF(acontext->local = Py_None);
	    Py_INCREF(acontext->remote = Py_None);
	}
    }
    return (acontext);

}   /* pk_new_AuthenticationContext() */


static PyObject*
get_krbhst(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    PyObject *list = NULL;
    char **realm_names = NULL;
    krb5_data realm;

    if (PyArg_ParseTuple(args, "s#:get_krbhst", &realm.data, &realm.length)) {
	krb5_error_code rc;

	rc = krb5_get_krbhst(pk_context, &realm, &realm_names);
	if (rc)
	    pk_error(rc);
	else {
	    int num = 0;
	    int curr = 0;

	    while (realm_names[num])
		++num;
	    list = PyList_New(num);
	    if (list == NULL) {
		goto finally;
	    }
	    for (; curr < num; ++curr) {
		PyObject *str = PyString_FromString(realm_names[curr]);

		if (str == NULL)
		    goto finally;
		PyList_SetItem(list, curr, str);
	    }
	    result = list;
	    list = NULL;
	}
    }
  finally:
    if (realm_names != NULL)
	krb5_free_krbhst(pk_context, realm_names);
    Py_XDECREF(list);

    return (result);

}   /* get_krbhst() */


static PyObject*
get_default_realm(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;

    if (PyArg_ParseTuple(args, ":get_default_realm")) {
	char *realm = NULL;
	krb5_error_code rc;

	rc = krb5_get_default_realm(pk_context, &realm);
	if (rc)
	    pk_error(rc);
	else {
	    result = PyString_FromString(realm);
	    free(realm);
	}
    }
    return (result);

}   /* get_default_realm() */


static PyMethodDef functions[] = {
    {"get_default_realm",	get_default_realm,	METH_VARARGS},
    {"get_krbhst",		get_krbhst,		METH_VARARGS},
    {"get_principal",		get_principal,		METH_VARARGS},
    {"get_login_principal",	get_login_principal,	METH_VARARGS},
    {"get_svc_principal",	get_svc_principal,	METH_VARARGS},
    {"net_read",		tcp_net_read,	        METH_VARARGS},
    {"net_write",		tcp_net_write,	        METH_VARARGS},

    /* tcp constructors */
    {"recv_auth",		pk_recv_auth,		METH_VARARGS},
    {"send_auth",		pk_send_auth,		METH_VARARGS},

    /* udp constructors */
    {"mk_req",			pk_mk_req,		METH_VARARGS},
    {"rd_req",			pk_rd_req,		METH_VARARGS},

    {NULL, NULL}
};


static void
addint(dict, name, value)
    PyObject *dict;
    char *name;
    long value;
{
    PyDict_SetItemString(dict, name, PyInt_FromLong(value));

}   /* addint() */


void
initkrb5()
{
    PyObject *module = Py_InitModule("krb5", functions);

    if (module != NULL) {
	PyObject *dict;
	krb5_error_code rc;

	empty_args_list = PyTuple_New(0);

	/* define and store the exception first, so that it can be raised if
	 * needed.
	 */
	dict = PyModule_GetDict(module);
#if PYTHON_API_VERSION > 1006
	/* Python 1.5 & newer */
	pk_exception = PyErr_NewException("krb5.Krb5Error", NULL, NULL);
#else
	/* Python 1.4 & older */
	pk_exception = PyString_FromString("krb5.Krb5Error");
#endif
	PyDict_SetItemString(dict, "Krb5Error", pk_exception);

	/* Store other misc. values. */
	PyDict_SetItemString(dict, "PrincipalType",
			     (PyObject *)&pk_PrincipalType);
	PyDict_SetItemString(dict, "UDPAuthenticationContextType",
			     (PyObject *)&pk_UDPAuthenticationContextType);
	PyDict_SetItemString(dict, "TCPAuthenticationContextType",
			     (PyObject *)&pk_UDPAuthenticationContextType);

	PyDict_SetItemString(dict, "BLOCK_SAFE",
			     PyString_FromString("safe"));
	PyDict_SetItemString(dict, "BLOCK_PRIVATE",
			     PyString_FromString("priv"));

	/* Kerberos constants */
	addint(dict, "AP_OPTS_USE_SESSION_KEY", AP_OPTS_USE_SESSION_KEY);
	addint(dict, "AP_OPTS_MUTUAL_REQUIRED", AP_OPTS_MUTUAL_REQUIRED);
	addint(dict, "KRB5_AUTH_CONTEXT_DO_TIME", KRB5_AUTH_CONTEXT_DO_TIME);
	addint(dict, "KRB5_AUTH_CONTEXT_DO_SEQUENCE",
	       KRB5_AUTH_CONTEXT_DO_SEQUENCE);
	addint(dict, "KRB5_PROG_SUMTYPE_NOSUPP", KRB5_PROG_SUMTYPE_NOSUPP);
	addint(dict, "KRB5_SENDAUTH_REJECTED", KRB5_SENDAUTH_REJECTED);
	addint(dict, "KRB5_FCC_NOFILE", KRB5_FCC_NOFILE);
	addint(dict, "KRB5_PARSE_MALFORMED", KRB5_PARSE_MALFORMED);
	addint(dict, "KRB5KRB_AP_ERR_BADADDR", KRB5KRB_AP_ERR_BADADDR);
	addint(dict, "KRB5KRB_AP_ERR_BADORDER", KRB5KRB_AP_ERR_BADORDER);
	addint(dict, "KRB5KRB_AP_ERR_BAD_INTEGRITY",
	       KRB5KRB_AP_ERR_BAD_INTEGRITY);
	addint(dict, "KRB5KRB_AP_ERR_TKT_EXPIRED", KRB5KRB_AP_ERR_TKT_EXPIRED);
	addint(dict, "KRB5KRB_AP_ERR_TKT_NYV", KRB5KRB_AP_ERR_TKT_NYV);
	addint(dict, "KRB5_CONFIG_CANTOPEN", KRB5_CONFIG_CANTOPEN);

	/* initialize Kerberos */
	rc = krb5_init_context(&pk_context);
	if (rc) {
	    pk_error(rc);
	    return;
	}
	rc = krb5_cc_default(pk_context, &pk_ccache);
	if (rc) {
	    pk_error(rc);
	    return;
	}
    }
}   /* initkrb5() */
