/*  Internal header for the krb5 Python extension module.
 *
 *  $Revision: 1.30 $
 *  $Source: /projects/cvsroot/goodies/krb5module/krb5py.h,v $
 *
 *  Written by Fred L. Drake, Jr.  <fdrake@cnri.reston.va.us>
 *  and Roger E. Masse		   <rmasse@cnri.reston.va.us>
 */

#include "Python.h"

#include "krb5.h"			/* Kerberos interfaces */
#include "com_err.h"			/* error message retrieval */

#define TCPVERSION "KRB5_TCP_Python_v1.0"

#ifdef MS_WINDOWS
#include <winsock.h>
#else
#include "myselect.h"                   /* Also includes mytime.h */
#define SOCKET int
#endif

#include <netinet/in.h>			/* struct sockaddr_in */


#if PYTHON_API_VERSION < 1007
/* Python 1.4 and older */
#undef PyString_AS_STRING
#define PyString_AS_STRING(op) (((PyStringObject *)(op))->ob_sval)
#undef PyString_GET_SIZE
#define PyString_GET_SIZE(op)  (((PyStringObject *)(op))->ob_size)
#endif


/*  These declarations are needed since they are missing from the krb5.h file.
 */
extern krb5_error_code
krb5_gen_replay_name Py_PROTO((/* IN */
			       krb5_context context,
			       const krb5_address *inaddr,
			       const char *uniq,
			       /* OUT */
			       krb5_pointer fillin));

extern int
krb5_net_read Py_PROTO((/* IN */
			krb5_context context,
			int fd,
			/* OUT */
			char *buf,
			/* IN */
			int len));

extern int
krb5_net_write Py_PROTO((/* IN */
			 krb5_context context,
			 int fd,
			 const char *buf,
			 int len));

/*  This structure represents a krb5_auth_context to the Python layer.
 *  The embedded ticket is used primarily to support the unparse_name()
 *  method.  Both the tcp_auth_context and udp_auth_context are maintained
 *  in these structures.
 */
typedef struct _AuthenticationContext {
    PyObject_HEAD
    PyObject *local;			/* local principal */
    PyObject *remote;			/* remote principal */
    krb5_auth_context auth_context;
    krb5_ticket *ticket;
    krb5_rcache rcache;
    int have_addresses;
    int have_rcache;
    /* additional stuff for TCP variant */
    PyObject *sock;
    SOCKET fd;
} AuthenticationContext;


/*  This structure represents a principle (user, server) in the kerberos
 *  environment.  The service and host attributes are made available for
 *  convenience.
 */
typedef struct _Principal {
    PyObject_HEAD
    PyObject *service;			/* service passed to contructor */
    PyObject *host;			/* hostname passed to constructor */
    PyObject *name;			/* unparsed name, prepared on demand */
    krb5_principal principal;
} Principal;

/* Differeent types of block I/O operations defined in krb5tcp.c */
#define PK_BLOCK_NONE 0
#define PK_BLOCK_SAFE 1
#define PK_BLOCK_PRIV 2

typedef struct _pk_block_info {
    short xmitlen;               /* len we're about to write */
    short type;                  /* One of the above PK_BLOCK flags */
} pk_block_info;

/*  The TCP auth_context constructors.
 *
 *  These are implemented in krb5tcp.c.
 */
extern PyObject*
pk_recv_auth Py_PROTO((PyObject *self, PyObject *args));

extern PyObject*
pk_send_auth Py_PROTO((PyObject *self, PyObject *args));

/*  The UDP auth_context constructors.
 *
 *  These are implemented in krb5udp.c.
 */
extern PyObject*
pk_mk_req Py_PROTO((PyObject *self, PyObject *args));

extern PyObject*
pk_rd_req Py_PROTO((PyObject *self, PyObject *args));


/*  These are the data encoding/decoding routines; they are shared by
 *  the TCP and UDP AuthenticationContext objects.  (The "ac" in the
 *  names stands for "AuthenticationContext".)
 */
extern PyObject*
pk_ac_mk_priv Py_PROTO((AuthenticationContext *self, PyObject *args));

extern PyObject*
pk_ac_mk_safe Py_PROTO((AuthenticationContext *self, PyObject *args));

extern PyObject*
pk_ac_rd_priv Py_PROTO((AuthenticationContext *self, PyObject *args));

extern PyObject*
pk_ac_rd_safe Py_PROTO((AuthenticationContext *self, PyObject *args));


/*  Misc. AuthenticationContext methods.
 */
extern PyObject*
pk_ac_getflags Py_PROTO((AuthenticationContext *self, PyObject *args));

extern PyObject*
pk_ac_setflags Py_PROTO((AuthenticationContext *self, PyObject *args));


/*  Internal operations on AuthenticationContext objects; these are not
 *  methods.
 */
extern int
pk_ac_set_addresses Py_PROTO((AuthenticationContext *self,
			      struct sockaddr_in *local,
			      struct sockaddr_in *remote));

extern int
pk_ac_set_rcache Py_PROTO((AuthenticationContext *self,
			   struct sockaddr_in *remote,
			   const char *cltname));

/*  This calls both of the above functions if necessary; only valid for the
 *  TCP version of the AuthenticationContext objects.
 */
extern int
pk_tcp_setup_internals Py_PROTO((AuthenticationContext *self));


/*  Internal utility functions to raise an exception.
 */
extern void
pk_error Py_PROTO((krb5_error_code rc));

extern void
pk_var_error Py_PROTO((krb5_error_code rc, const char *fmt, ...));

/*  The actual exception object.
 */
extern PyObject *pk_exception;


/*  Internal utility functions.
 *
 *  pk_socket_error() raises a socket.error exception with the appropriate
 *  message.
 *
 *  pk_parse_address() returns the IP number for a host name in network byte
 *  order.  Input name is anything accepted for Python socket host names,
 *  except '<broadcast>'.  (Which is not needed here.)
 */
extern void
pk_socket_error Py_PROTO(());

extern int
pk_check_socket Py_PROTO((PyObject *sock));

extern int
pk_conv_fd Py_PROTO((PyObject *obj, void *result));

extern unsigned long
pk_parse_address Py_PROTO((const char *address));


/*  Internal utility to create a new, partially initialized Principal object.
 *
 *  Each parameter may be either a Python string object or None.
 */
extern Principal*
pk_new_Principal Py_PROTO((PyObject *hostname, PyObject *service));


/*  Internal utility to create a new AuthenticationContext; the type passed
 *  in should be either UDPAuthenticationContextType or
 *  TCPAuthenticationContextType, depending on how it was created.
 */
extern AuthenticationContext*
pk_new_AuthenticationContext Py_PROTO((PyTypeObject *type));


/*  Internal data shared among the implementation files:
 *
 *  pk_context is the per-process krb5_context object.
 *  pk_ccache is the per-process credentials cache (should allow multiple
 *  caches for a fully general Kerberos 5 module).
 */
extern krb5_context pk_context;
extern krb5_ccache pk_ccache;

/*  Type objects for the types defined by the krb5 module.
 */
extern PyTypeObject pk_PrincipalType;
extern PyTypeObject pk_UDPAuthenticationContextType;
extern PyTypeObject pk_TCPAuthenticationContextType;
