/*  Python extension module for Kerberos 5 interfaces -- TCP auth_context.
 *
 *  Written by Roger E. Masse      <rmasse@CNRI.Reston.Va.US>
 *         and Fred L. Drake, Jr.  <fdrake@cnri.reston.va.us>
 */

#include "krb5py.h"			/* common header file */
#include "structmember.h"		/* Python attribute lookup */

#include <sys/socket.h>                 /* struct sockaddr */
#include <errno.h>                      /* errno global */
#include <netdb.h>                      /* gethostbyname */

char *btype2str(type)
    short type;
{
    switch (type) {
        case PK_BLOCK_NONE:
	    return "";
        case PK_BLOCK_SAFE:
	    return "safe";
        case PK_BLOCK_PRIV:
	    return "priv";
        default:
	    pk_var_error(0, "Unknown block type (%d)", type);
	    return "unknown"; 
    }
}   /* btype2str() */


static PyObject*
pk_tcp_block_write(self, userdata, mk_func, type)
    AuthenticationContext *self;
    krb5_data userdata;
    krb5_error_code (*mk_func)();
    short type;
{
    PyObject *result = NULL;
    krb5_data outbuf;
    krb5_error_code rc;
    pk_block_info xmitblock;

    outbuf.data = NULL;
    outbuf.length = 0;

    /* Call optional mk_func and package message accordingly */
    if (mk_func) {
	if (pk_tcp_setup_internals(self) == -1)
	    goto finally;
        rc = mk_func(pk_context, self->auth_context, &userdata, &outbuf, NULL);
        if (rc) {
	    char *p = btype2str(type);
	    printf("Type = %d, btype2str(type)=\"%s\"\n", type, p);
            pk_var_error(rc, "while making message."); /* GROT */
            goto finally;
        }
    }
    else
        outbuf = userdata;

    /* Use net_write to send the length of the data and the type of
     * the stream... i.e. is it priv, safe, or none.
     */
    xmitblock.xmitlen = htons(outbuf.length);
    xmitblock.type = htons(type);

    rc = krb5_net_write(pk_context, self->fd,
			(char *)&xmitblock, sizeof(xmitblock));
    if (rc < 0) {
	pk_socket_error();
        goto finally;
    }
    rc = krb5_net_write(pk_context, self->fd,
			(char *)outbuf.data, outbuf.length);
    if (rc < 0) {
	pk_socket_error();
        goto finally;
    }
    Py_INCREF(result = Py_None);

  finally:
    if (mk_func)
        free(outbuf.data);

    return (result);   	

}   /* pk_tcp_block_write() */


static PyObject* 
pk_tcp_block_read(self)
    AuthenticationContext *self;
{
    PyObject *result = NULL;
    short write_type;      
    krb5_data outbuf;
    krb5_data recvdata;
    krb5_error_code rc;
    pk_block_info recvblock;
    krb5_error_code (*rd_func)();

    if ((rc = krb5_net_read(pk_context, self->fd, (char *)&recvblock,
			    sizeof(recvblock))) <= 0) {
	if (rc == 0)
	    errno = ECONNABORTED;
	pk_socket_error();
	goto finally;
    }

    /* get the length we should allocate and the packet type for decoding */
    recvdata.length = ntohs(recvblock.xmitlen);
    write_type = ntohs(recvblock.type);

    /* allocate enough space for a NUL byte */
    if (!(recvdata.data = (char *)malloc(recvdata.length + 1))) {
	PyErr_NoMemory();
        goto finally;
    }
    if ((rc = krb5_net_read(pk_context, self->fd, (char *)recvdata.data,
			    recvdata.length)) <= 0) {
        if (rc == 0)
	    errno = ECONNABORTED;
	pk_socket_error();
	goto finally;
    }
    recvdata.data[recvdata.length] = '\0';

    /* determine weither any special encapsulation was used */
    switch(write_type) {
        case PK_BLOCK_NONE:
	    rd_func = NULL;
	    break;
        case PK_BLOCK_SAFE:
	    rd_func = krb5_rd_safe;
	    break;
        case PK_BLOCK_PRIV:
            rd_func = krb5_rd_priv;
	    break;
        default:
	    pk_var_error(0, "Unknown block type obtained from writer");
	    /*  btype2str(write_type)); */
	    goto finally;
    }

    /* call optional rd_func to check for special unpackaging */
    if (rd_func) {
	if (pk_tcp_setup_internals(self) == -1)
	    goto finally;
	outbuf.data = NULL;
	outbuf.length = 0;
        rc = rd_func(pk_context, self->auth_context, &recvdata, &outbuf, NULL);
        if (rc) {
	    pk_var_error(rc, "while reading message.");
	    /* btype2str(write_type)); */
	    goto finally;
        }
    }
    else
        outbuf = recvdata;

    result = Py_BuildValue("ss#", btype2str(write_type),
			   outbuf.data, outbuf.length);
    if (rd_func)
	free(outbuf.data);
    /* krb5_free_data(pk_context, &outbuf); */

  finally:
    free(recvdata.data);

    return (result);

}   /* tcp_ac_block_read() */


static PyObject*
tcp_ac_block_write(self, args)
    AuthenticationContext *self;
    PyObject *args;
{  
    krb5_data userdata;
    if (PyArg_ParseTuple(args, "s#:block_write",
			 &userdata.data, &userdata.length))
        return pk_tcp_block_write(self, userdata, NULL, PK_BLOCK_NONE);
    return NULL;

}   /* tcp_ac_block_write() */

static PyObject*
tcp_ac_block_write_safe(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    krb5_data userdata;
    if (PyArg_ParseTuple(args, "s#:block_write_safe",
			 &userdata.data, &userdata.length))
        return pk_tcp_block_write(self, userdata, krb5_mk_safe, PK_BLOCK_SAFE);
    return NULL;

}   /* tcp_ac_block_write_safe() */

static PyObject*
tcp_ac_block_write_priv(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    krb5_data userdata;
    if (PyArg_ParseTuple(args, "s#:block_write_priv",
			 &userdata.data, &userdata.length))
        return pk_tcp_block_write(self, userdata, krb5_mk_priv, PK_BLOCK_PRIV);
    return NULL;

}   /* tcp_ac_block_write_priv() */

static PyObject*
tcp_ac_block_read(self, args)
    AuthenticationContext *self;
    PyObject *args;
{
    if (PyArg_ParseTuple(args, ":block_read"))
        return pk_tcp_block_read(self);
    return NULL;    

}   /* tcp_ac_block_read() */


#define OFF(x) offsetof(AuthenticationContext, x)

static struct memberlist
ac_memberlist[] = {
    {"remote",		T_OBJECT,	OFF(remote),		RO},
    {"local",		T_OBJECT,	OFF(local),		RO},

    {NULL}
};


static PyMethodDef
tcp_ac_methods[] = {

    /* safe packet handling - encrypted checksum */
    {"rd_safe",		 (PyCFunction)pk_ac_rd_safe,	METH_VARARGS,
     "Decode a message with an encypted checksum."},
    {"mk_safe",		 (PyCFunction)pk_ac_mk_safe,	METH_VARARGS,
     "Encode a message with an encypted checksum."},
    {"block_write_safe", tcp_ac_block_write_safe,	METH_VARARGS},

    /* encrypted packet handling */
    {"rd_priv",		 (PyCFunction)pk_ac_rd_priv,	METH_VARARGS,
     "Decode an encrypted message."},
    {"mk_priv",		 (PyCFunction)pk_ac_mk_priv,	METH_VARARGS,
     "Encode an encrypted message."},
    {"block_write_priv", tcp_ac_block_write_priv,	METH_VARARGS},
    
    /* ordinary packet handling */
    {"block_read",	 tcp_ac_block_read,	        METH_VARARGS},
    {"block_write",	 tcp_ac_block_write,	        METH_VARARGS},

    /* other */
    {"getflags",	(PyCFunction)pk_ac_getflags,	METH_VARARGS,
     "Get flags controlling use & return of timestamps and sequence numbers."},
    {"setflags",	(PyCFunction)pk_ac_setflags,	METH_VARARGS,
     "Set flags controlling use & return of timestamps and sequence numbers."},

    {NULL}
};

static PyObject*
tcp_ac_getattr(self, name)
    PyObject *self;
    char *name;
{
    PyObject *result = PyMember_Get((char *)self, ac_memberlist, name);

    if (result == NULL) {
	PyErr_Clear();
	result = Py_FindMethod(tcp_ac_methods, self, name);
    }
    return (result);

}   /* tcp_ac_getattr() */


PyTypeObject
pk_TCPAuthenticationContextType = {
    PyObject_HEAD_INIT(NULL)
    0,
    "TCPAuthenticationContext",		/* tp_name		*/
    sizeof(AuthenticationContext),	/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    0,					/* tp_dealloc		*/
    0,					/* tp_print		*/
    tcp_ac_getattr,	       		/* tp_getattr		*/
    0,					/* tp_setattr		*/
    0,					/* tp_compare		*/
    0,	       		                /* tp_repr		*/
    0,					/* tp_as_number		*/
    0,					/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    0,					/* tp_getattro		*/
    0,					/* tp_setattro		*/
    0,					/* tp_as_buffer		*/

    /* Space for future expansion */
    0,					/* tp_xxx4		*/

    /* __doc__ */
    "Kerberos authentication context for TCP connections."
};


/* constructors for TCPAuthenticationContext */

PyObject *
pk_send_auth(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;

    PyObject *sock = NULL;
    AuthenticationContext *acontext = NULL;
    Principal *c_princ;
    Principal *s_princ;
    krb5_flags ap_req_options = 0;
    krb5_data cksum_data;
    krb5_error_code rc;
    krb5_error *err_ret;
    krb5_ap_rep_enc_part *rep_ret;

    if (PyArg_ParseTuple(args, "OO!O!|i:send_auth", &sock,
			 &pk_PrincipalType, &c_princ,
		         &pk_PrincipalType, &s_princ,
			 &ap_req_options)) {
	SOCKET s = pk_check_socket(sock);

	if (s == -1)
	    goto finally;

        /* pull the server hostname and service out of the Principal */
	if (s_princ->host == Py_None) {
	    PyErr_SetString(PyExc_ValueError,
			    "Malformed server principal, missing hostname.");
	    goto finally;
        }
	if (s_princ->service == Py_None) {
	    PyErr_SetString(PyExc_ValueError,
			    "Malformed server principal, missing service.");
	    goto finally;
        }
	
        /* check for CRC32 checksum support */
        if (!valid_cksumtype(CKSUMTYPE_CRC32)) {
	    pk_var_error(KRB5_PROG_SUMTYPE_NOSUPP, "while using CRC-32");
	    goto finally;
        }

        /* A Pythonized copy of a (extended) krb5 authcontext is */
        /* what we building to be returned */
        acontext =
	    pk_new_AuthenticationContext(&pk_TCPAuthenticationContextType);
        if (acontext == NULL)
	    goto finally;

        /* make up some ckecksum data using server hostname */
	if (!PyString_Check(s_princ->host)) {
	    PyErr_SetString(PyExc_ValueError,
			    "host attribute of the server principal not set");
	    goto finally;
	}
        cksum_data.data = PyString_AS_STRING(s_princ->host);
	cksum_data.length = PyString_GET_SIZE(s_princ->host);

        /* send the auth request */
        rc = krb5_sendauth(pk_context, &(acontext->auth_context),
			   (krb5_pointer) &s, TCPVERSION,
			   c_princ->principal, s_princ->principal,
			   ap_req_options,
			   &cksum_data,
			   0,		/* no creds, use ccache instead */
			   pk_ccache, &err_ret, &rep_ret, NULL);

        /* check return for errors */
        if (rc && rc != KRB5_SENDAUTH_REJECTED) {
	    pk_error(rc);
	    goto finally;
        }
        if (rc == KRB5_SENDAUTH_REJECTED) {
	    /* got an error */
	    pk_var_error(rc, "(%.*s)",
			 err_ret->text.length, err_ret->text.data);
        }
	else if (rep_ret) {
	    PyObject *temp = acontext->remote;

	    Py_INCREF(acontext->remote = (PyObject *)s_princ);
	    Py_XDECREF(temp);
	    temp = acontext->local;
	    Py_INCREF(acontext->local = (PyObject *)c_princ);
	    Py_XDECREF(temp);
	    if (!PyInt_Check(sock))
		Py_INCREF(acontext->sock = sock);
	    acontext->fd = s;

	    if (pk_tcp_setup_internals(
		    (AuthenticationContext *)acontext) == -1) {
		goto finally;
	    }

	    /* set result */
	    Py_INCREF(result = (PyObject *)acontext);
	}
        else
	    pk_var_error(0, "no error or reply from sendauth.");
    }
  finally:
    Py_XDECREF(acontext);

    return (result);

}   /* pk_send_auth() */


PyObject *
pk_recv_auth(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;

    PyObject *sock;
    AuthenticationContext *acontext = NULL;
    Principal *s_princ;
    krb5_error_code rc;

    if (PyArg_ParseTuple(args, "OO!:recv_auth", &sock,
			 &pk_PrincipalType, &s_princ)) {
	SOCKET s = pk_check_socket(sock);

	if (s == -1)
	    goto finally;

        /* pull service string from Principal */
	if (s_princ->service == Py_None) {
	    PyErr_SetString(PyExc_ValueError,
			    "principal does not represent a service");
	    goto finally;
        }

        /* A Pythonized copy of a (extended) krb5 authcontext is */
        /* what we are building to be returned */
        acontext =
	    pk_new_AuthenticationContext(&pk_TCPAuthenticationContextType);
        if (acontext == NULL)
	    goto finally;

	acontext->fd = s;
	if (!PyInt_Check(sock))
	    Py_INCREF(acontext->sock = sock);

        /* Read authorization message */
	rc = krb5_recvauth(pk_context, &acontext->auth_context,
			   (krb5_pointer)&s,
			   TCPVERSION, s_princ->principal,
			   0,		/* no flags */
			   NULL,	/* default keytab is NULL */
			   &acontext->ticket);
        if (rc) {
            pk_error(rc);
	    goto finally;
	}
	Py_DECREF(acontext->local);
	Py_INCREF(acontext->local = (PyObject *)s_princ);
	{
	    Principal *remote = pk_new_Principal(Py_None, Py_None);
	    PyObject *temp;
	    krb5_error_code rc;

	    if (remote == NULL)
		goto finally;
	    rc = krb5_copy_principal(pk_context,
				     acontext->ticket->enc_part2->client,
				     &remote->principal);
	    if (rc) {
		pk_error(rc);
		Py_DECREF(remote);
		goto finally;
	    }
	    temp = acontext->remote;
	    Py_INCREF(acontext->remote = (PyObject *)remote);
	    Py_DECREF(temp);
	}
	/* set result */
	Py_INCREF(result = (PyObject *)acontext);
    }
  finally:
    Py_XDECREF(acontext);

    return (result);

}   /* pk_recv_auth() */
