/*  Python extension module for Kerberos 5 interfaces -- UDP auth_context.
 *
 *  Written by Fred L. Drake, Jr.  <fdrake@cnri.reston.va.us>
 */


#include "krb5py.h"			/* common header file */
#include "structmember.h"		/* Python attribute lookup */


static int
udp_setup_internals(acontext, hostaddr, port, rhostaddr, rport)
    AuthenticationContext *acontext;
    char *hostaddr;
    int port;
    char *rhostaddr;
    int rport;
{
    int result = -1;
    struct sockaddr_in address;
    struct sockaddr_in raddress;

    /* fill in the local address structure */
    address.sin_port = htons((short)port);
    address.sin_addr.s_addr = pk_parse_address(hostaddr);
    if (address.sin_addr.s_addr == 0)
	goto finally;

    /* fill in the remote address structure */
    raddress.sin_port = htons((short)rport);
    raddress.sin_addr.s_addr = pk_parse_address(rhostaddr);
    if (raddress.sin_addr.s_addr == 0)
	goto finally;

    /* set up Kerberos */
    if (pk_ac_set_addresses(acontext, &address, &raddress) == -1)
	goto finally;
    result = pk_ac_set_rcache(acontext, &address, "_python_krb5_");

  finally:
    return (result);

}   /* udp_setup_internals() */


#define OFF(x) offsetof(AuthenticationContext, x)

static struct memberlist
ac_memberlist[] = {
    {"remote",		T_OBJECT,	OFF(remote),		RO},
    {"local",		T_OBJECT,	OFF(local),		RO},

    {NULL}
};


static PyMethodDef
ac_methods[] = {
    /* safe packet handling */
    {"rd_safe",		(PyCFunction)pk_ac_rd_safe,	METH_VARARGS,
     "Decode a message with an encypted checksum."},
    {"mk_safe",		(PyCFunction)pk_ac_mk_safe,	METH_VARARGS,
     "Encode a message with an encypted checksum."},

    /* encrypted packet handling */
    {"rd_priv",		(PyCFunction)pk_ac_rd_priv,	METH_VARARGS,
     "Decode an encrypted message."},
    {"mk_priv",		(PyCFunction)pk_ac_mk_priv,	METH_VARARGS,
     "Encode an encrypted message."},

    /* other */
    {"getflags",	(PyCFunction)pk_ac_getflags,	METH_VARARGS,
     "Get flags controlling use & return of timestamps and sequence numbers."},
    {"setflags",	(PyCFunction)pk_ac_setflags,	METH_VARARGS,
     "Set flags controlling use & return of timestamps and sequence numbers."},

    {NULL}

};   /* ac_methods[] */


static PyObject*
ac_getattr(self, name)
    PyObject *self;
    char *name;
{
    PyObject *result = PyMember_Get((char *)self, ac_memberlist, name);

    if (result == NULL) {
	PyErr_Clear();
	result = Py_FindMethod(ac_methods, self, name);
    }
    return (result);

}   /* ac_getattr() */


PyTypeObject
pk_UDPAuthenticationContextType = {
    PyObject_HEAD_INIT(NULL)
    0,
    "krb5.UDPAuthenticationContext",	/* tp_name		*/
    sizeof(AuthenticationContext),	/* tp_basicsize		*/
    0,					/* tp_itemsize		*/
    0,					/* tp_dealloc		*/
    0,					/* tp_print		*/
    ac_getattr,				/* tp_getattr		*/
    0,					/* tp_setattr		*/
    0,					/* tp_compare		*/
    0,					/* tp_repr		*/
    0,					/* tp_as_number		*/
    0,					/* tp_as_sequence	*/
    0,					/* tp_as_mapping	*/
    0,					/* tp_hash		*/
    0,					/* tp_call		*/
    0,					/* tp_str		*/
    0,					/* tp_getattro		*/
    0,					/* tp_setattro		*/

    /* Functions to access object as input/output buffer */
    0,					/* tp_as_buffer		*/

    /* Space for future expansion */
    0,					/* tp_xxx4		*/

    /* __doc__ */
    "Kerberos authentication context for UDP connections."
};


PyObject*
pk_mk_req(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    AuthenticationContext *acontext = NULL;
    Principal *princ;
    krb5_data in_data;
    krb5_data outbuf;
    char *hostaddr;
    int port;
    char *rhostaddr;
    int rport;
    krb5_flags ap_req_options = 0;

    outbuf.data = NULL;
    if (PyArg_ParseTuple(args, "O!(si)(si)|i:mk_req",
			 &pk_PrincipalType, &princ,
			 /* this is not used since we don't understand why;
			  * to change this, tell us why it should be
			  * exposed: */
			 /* "s#" &in_data.data, &in_data.length, */
			 &hostaddr, &port,
			 &rhostaddr, &rport,
			 &ap_req_options)) {
	krb5_error_code rc;
	char *service;
	char *hostname;

	/*  This is done in the sample client, but I don't know why.
	 */
	if (!valid_cksumtype(CKSUMTYPE_CRC32)) {
	    pk_var_error(0, "CRC32 is an invalid check-sum type");
	    goto finally;
	}
	if (princ->host == Py_None) {
	    PyErr_SetString(PyExc_ValueError,
			    "Malformed server principal, missing hostname.");
	    goto finally;
        }
	if (princ->service == Py_None) {
	    PyErr_SetString(PyExc_ValueError,
			    "Malformed server principal, missing service.");
	    goto finally;
        }
	/* do this to have data without having to get
	 * it from the Python level: */
	in_data.data = PyString_AS_STRING(princ->host);
	in_data.length = PyString_GET_SIZE(princ->host);
	acontext = pk_new_AuthenticationContext(
	    &pk_UDPAuthenticationContextType);
	if (acontext == NULL)
	    goto finally;
	hostname = PyString_AS_STRING(princ->host);
	service = PyString_AS_STRING(princ->service);
	rc = krb5_mk_req(pk_context, &acontext->auth_context,
			 ap_req_options, service, hostname, &in_data,
			 pk_ccache, &outbuf);
	if (rc) {
	    pk_error(rc);
	    goto finally;
	}
	if (udp_setup_internals(acontext, hostaddr, port, rhostaddr, rport))
	    goto finally;
	{
	    PyObject *temp = acontext->remote;

	    Py_INCREF(acontext->remote = (PyObject *)princ);
	    Py_DECREF(temp);
	    result = Py_BuildValue(
		"Os#", acontext, outbuf.data, outbuf.length);
	}
    }
  finally:
    free(outbuf.data);
    Py_XDECREF(acontext);

    return (result);

}   /* pk_mk_req() */


PyObject*
pk_rd_req(self, args)
    PyObject *self;
    PyObject *args;
{
    PyObject *result = NULL;
    Principal *princ;
    krb5_data inbuf;
    char *hostaddr;
    int port;
    char *rhostaddr;
    int rport;
    AuthenticationContext *acontext = NULL;

    if (PyArg_ParseTuple(args, "O!s#(si)(si):rd_req",
			 &pk_PrincipalType, &princ,
			 &inbuf.data, &inbuf.length,
			 &hostaddr, &port,
			 &rhostaddr, &rport)) {
	krb5_error_code rc;

	acontext = pk_new_AuthenticationContext(
	    &pk_UDPAuthenticationContextType);
	if (acontext == NULL)
	    goto finally;
	rc = krb5_rd_req(pk_context, &acontext->auth_context,
			 &inbuf, princ->principal, NULL,
			 NULL, &acontext->ticket);
	if (rc) {
	    pk_error(rc);
	    goto finally;
	}
	{
	    PyObject *temp;
	    Principal *remote = pk_new_Principal(Py_None, Py_None);

	    if (remote != NULL) {
		rc = krb5_copy_principal(
		    pk_context, acontext->ticket->enc_part2->client,
		    &remote->principal);
		if (rc) {
		    pk_error(rc);
		    Py_DECREF(remote);
		}
		else {
		    temp = acontext->remote;
		    acontext->remote = (PyObject *)remote;
		    Py_DECREF(temp);
		}
	    }
	    temp = acontext->local;
	    Py_INCREF(acontext->local = (PyObject *)princ);
	    Py_DECREF(temp);
	}
	if (!udp_setup_internals(acontext, hostaddr, port, rhostaddr, rport))
	    Py_INCREF(result = (PyObject *)acontext);
    }
  finally:
    Py_XDECREF(acontext);

    return (result);

}   /* pk_rd_req() */
